/*
 * linux/arch/arm/kernel/ds2400.c
 *
 * Driver for Dallas "Silicon Serial Number" chip
 * Copyright (C) 1999 Phil Blundell
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * Refer to <http://www.dalsemi.com/DocControl/PDFs/2401.pdf> for
 * programming details of the Dallas part.  In Acorn machines the
 * one-bit bus is hung off IOC pin C[3].  Redefining the OWB_* macros
 * should suffice to adapt to new hardware.  
 */

#include <linux/delay.h>
#include <linux/sched.h>
#include <linux/init.h>

#include <asm/system.h>
#include <asm/hardware.h>
#include <asm/io.h>
#include <asm/ioc.h>

extern unsigned long long system_serial_num;		/* from setup.c */

#define OWB_TX(n)	outb(n ? 0xDF : 0xD7, IOC_CONTROL)
#define OWB_RX()	((inb(IOC_CONTROL) & 0x8) ? 1 : 0)

/*
 * Send a reset on the one wire bus and wait for the presence pulse
 */

static int __init owb_reset(void)
{
	unsigned long j;
	OWB_TX(0);
	udelay(600);
	OWB_TX(1);
	j = jiffies;
	udelay(20);
	while (OWB_RX()) {
		if ((jiffies - j) > 2)
			return 1;
	}
	udelay(600);
	return 0;
}

/*
 * Send a bit on the one wire bus
 */

static inline void owb_writebit(int n)
{
	OWB_TX(0);
	udelay(10);
	if (n) {
		OWB_TX(1);
		udelay(80);
	} else {
		udelay(60);
		OWB_TX(1);
		udelay(10);
	}
}

/*
 * Receive a bit on the one wire bus
 */

static inline int owb_readbit(void)
{
	int r;
	OWB_TX(0);
	udelay(7);
	OWB_TX(1);
	udelay(7);
	r = OWB_RX();
	udelay(60);
	return r;
}

void __init ds2400_init(void)
{
	unsigned int i;
	unsigned long long d = 0;

	if (owb_reset()) {
		printk("DS2400 is toast; unable to read machine ID\n");
		return;
	}

	for (i = 0; i < 8; i++)
		owb_writebit(0xf & (1<<i));

	for (i = 0; i < 64; i++)
		d |= (owb_readbit() << i);

	/* Strip off the top two bytes, which are actually the family ID
	   and data CRC, to leave the 48-bit unique ID.  */
	system_serial_num = d & 0x0000ffffffffffffULL;
}
