//#define __KERNEL_SYSCALLS__
//#include <stdarg.h>

#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/smp.h>
#include <linux/smp_lock.h>
#include <linux/stddef.h>
#include <linux/unistd.h>
#include <linux/ptrace.h>
#include <linux/malloc.h>
#include <linux/vmalloc.h>
#include <linux/user.h>
#include <linux/a.out.h>
#include <linux/interrupt.h>
#include <linux/config.h>
#include <linux/delay.h>
#include <linux/reboot.h>
#include <linux/init.h>

#include <asm/uaccess.h>
#include <asm/system.h>
#include <asm/io.h>


/* Offsets to the Command and Status Registers, "CSRs".  All accesses
   must be longword instructions and quadword aligned. */
enum tulip_offsets
{
  CSR0 = 0, CSR1 = 0x08, CSR2 = 0x10, CSR3 = 0x18, CSR4 = 0x20, CSR5 = 0x28,
  CSR6 = 0x30, CSR7 = 0x38, CSR8 = 0x40, CSR9 = 0x48, CSR10 = 0x50, CSR11 =
    0x58,
  CSR12 = 0x60, CSR13 = 0x68, CSR14 = 0x70, CSR15 = 0x78
};

	//*
	//* Netwinder has memory mapped access to the 21143's CSRs,
	//* therefore we have to redefine register input and output.
	//* */

#undef outl(val,port)
#undef inl(port)

#define outl(val,port)	writel((val),(port))
#define inl(port)	readl((port))


/* MII transceiver control section.
   Read and write the MII registers using software-generated serial
   MDIO protocol.  See the MII specifications or DP83840A data sheet
   for details. */

/* The maximum data clock rate is 2.5 Mhz.  The minimum timing is usually
   met by back-to-back PCI I/O cycles, but we insert a delay to avoid
   "overclocking" issues or future 66Mhz PCI. */
#define mdio_delay() inl(mdio_addr)

/* Read and write the MII registers using software-generated serial
   MDIO protocol.  It is just different enough from the EEPROM protocol
   to not share code.  The maxium data clock rate is 2.5 Mhz. */
#define MDIO_SHIFT_CLK	0x10000
#define MDIO_DATA_WRITE0 0x00000
#define MDIO_DATA_WRITE1 0x20000
#define MDIO_ENB		0x00000	/* Ignore the 0x02000 databook setting. */
#define MDIO_ENB_IN		0x40000
#define MDIO_DATA_READ	0x80000

//static int mdio_read(struct device *dev, int phy_id, int location)
static int
mdio_read (int base_addr, int phy_id, int location)
{
  int i;
  int read_cmd = (0xf6 << 10) | (phy_id << 5) | location;
  int retval = 0;
  long mdio_addr = base_addr + CSR9;

  /* Establish sync by sending at least 32 logic ones. */
  for (i = 32; i >= 0; i--)
  {
    outl (MDIO_ENB | MDIO_DATA_WRITE1, mdio_addr);
    mdio_delay ();
    outl (MDIO_ENB | MDIO_DATA_WRITE1 | MDIO_SHIFT_CLK, mdio_addr);
    mdio_delay ();
  }
  /* Shift the read command bits out. */
  for (i = 15; i >= 0; i--)
  {
    int dataval = (read_cmd & (1 << i)) ? MDIO_DATA_WRITE1 : 0;

    outl (MDIO_ENB | dataval, mdio_addr);
    mdio_delay ();
    outl (MDIO_ENB | dataval | MDIO_SHIFT_CLK, mdio_addr);
    mdio_delay ();
  }
  /* Read the two transition, 16 data, and wire-idle bits. */
  for (i = 19; i > 0; i--)
  {
    outl (MDIO_ENB_IN, mdio_addr);
    mdio_delay ();
    retval = (retval << 1) | ((inl (mdio_addr) & MDIO_DATA_READ) ? 1 : 0);
    outl (MDIO_ENB_IN | MDIO_SHIFT_CLK, mdio_addr);
    mdio_delay ();
  }
  return (retval >> 1) & 0xffff;
}

static void
mdio_write (int base_addr, int phy_id, int location, int value)
{
  int i;
  int cmd = (0x5002 << 16) | (phy_id << 23) | (location << 18) | value;
  long mdio_addr = base_addr + CSR9;

  /* Establish sync by sending 32 logic ones. */
  for (i = 32; i >= 0; i--)
  {
    outl (MDIO_ENB | MDIO_DATA_WRITE1, mdio_addr);
    mdio_delay ();
    outl (MDIO_ENB | MDIO_DATA_WRITE1 | MDIO_SHIFT_CLK, mdio_addr);
    mdio_delay ();
  }
  /* Shift the command bits out. */
  for (i = 31; i >= 0; i--)
  {
    int dataval = (cmd & (1 << i)) ? MDIO_DATA_WRITE1 : 0;
    outl (MDIO_ENB | dataval, mdio_addr);
    mdio_delay ();
    outl (MDIO_ENB | dataval | MDIO_SHIFT_CLK, mdio_addr);
    mdio_delay ();
  }
  /* Clear out extra bits. */
  for (i = 2; i > 0; i--)
  {
    outl (MDIO_ENB_IN, mdio_addr);
    mdio_delay ();
    outl (MDIO_ENB_IN | MDIO_SHIFT_CLK, mdio_addr);
    mdio_delay ();
  }
  return;
}


static void 
tulip_pwrdn(void)
{
  unsigned int ioaddr;
  unsigned int phy;

  ioaddr = 0x00800000;		// mapped PCI memory address
  for (phy = 0; phy < 32; phy++)
  {
    if (mdio_read (ioaddr, phy, 0) == 0x3000)	//a powered on PHY?
    {
	mdio_write(ioaddr, phy, 0, 0x3800);
    }
  }
    //reuse int phy to power off the Tulip
    phy = *(unsigned long*)0xF8C04840;		// Tulip PCI config reg 40
    phy |= 0x80000000;			// power off bit
    *(unsigned long*)0xF8C04840 = phy;
}




#define BASE 0xE1800000

void w_grf(int reg, int data)
{
volatile unsigned char *p = (unsigned char *)BASE;	//that is where our regs are mapped in
    p[0x3CE] = reg;
    p[0x3CF] = data;
}

void w_crt(int reg, int data)
{
volatile unsigned char *p = (unsigned char *)BASE;	//that is where our regs are mapped in
    p[0x3D4] = reg;
    p[0x3D5] = data;
}

void w_seq(int reg, int data)
{
volatile unsigned char *p = (unsigned char *)BASE;	//that is where our regs are mapped in
    p[0x3C4] = reg;
    p[0x3C5] = data;
}

void netwinder_power_off(void)
{    
unsigned int	i;
volatile unsigned char *p = (unsigned char *)BASE;	//that is where our regs are mapped in
    //no more interrupts here... local CPU only...
    __cli();
    
    //shut down the hard disk NOW - send command SLEEP
    outb (0x99, 0x1F7);

    // give the HD a chance to report ready. Wait for BUSY bit to go away....
    i = 0x100000;
    while((inb(0x1F7) & 0x80) & i--) {};
    
    // shut down the video... 
    
//    i = p[0x3DA];		//dummy read to reset the flip flop
//    p[0x3C0] = 0x05;		//disable CRT out
//    p[0x3C0] = 0x00;

    w_seq(0x01,0x20);		//screen of
    w_seq(0x00,0x00);		//reset sequencer

    w_crt(0x17,0x33);		//HW reset, OC output control

    w_grf(0x5C,0x00);		//disable TV out - bit 5C<5>
    w_grf(0xDC,0x40);  		//disable video out, no CRT
    w_grf(0x56,0x1C);  		//enable extended RAMDAC 0x3C6, 0x3C7, 0x3C8
    p[0x3C6] = 0x55;		//DAC power down,bypass RAM, RAM power down
    p[0x3CF] = 0x00;		//stop access to 0x3C6

    w_grf(0x70,0x8B);		//reset sequencer bit<7>

    // turn off the PHY1 and Tulip
    tulip_pwrdn();
    
    // turn off all PCI devices

    // turn off the fan, turn off the LEDs
    gpio_modify_op(-1, 0);

    // turn off the Icache <bit 12> and Dcache <bit 2>
    __asm__ ("
    mrc p15, 0, r0, c1, c0
    bic r0, r0, #0x1000
    bic r0, r0, #0x0004
    mcr p15, 0, r0, c1, c0
    ");

#if 0
    *(volatile unsigned int*)0xFE00018C = 0xFFFFFFFF;	//disable all interrupts
    *(volatile unsigned int*)0xFE000308 = 0x0;		//disable timer1
    *(volatile unsigned int*)0xFE000328 = 0x0;		//disable timer2
    *(volatile unsigned int*)0xFE000348 = 0x0;		//disable timer3
    *(volatile unsigned int*)0xFE000368 = 0x0;		//disable timer4

    //commit harakiri: turn the ram off, while running with no cache
    *(volatile unsigned int*)0xFE000110 = 0;
    *(volatile unsigned int*)0xFE000114 = 0;
    *(volatile unsigned int*)0xFE000118 = 0;
    *(volatile unsigned int*)0xFE00011C = 0;
#endif    
    __sti();
    
}
