/*
 linux/arch/arm/kernel/nwtone.c

 Copyright 1998-1999 Rebel.com
 WoodyS & PatB

 Support for a little welcome sound, and
 for keyboard beeps

 IO     irq    dma     name
 220-f  12     3       soundblaster
 388-9                 adlib/opl
 
 279/a79/203           pnp control
 */
#include <linux/delay.h>
#include <linux/sched.h>
#include <linux/vt_kern.h>
#include <asm/system.h>
#include <asm/hardware.h>
#include <asm/io.h>

#define ADLIB_ADDR_PORT 0x388
#define ADLIB_DATA_PORT 0x389
#define BYTE char unsigned

typedef struct
{
    BYTE	flags1;
    BYTE	flags2;
    BYTE	ksl_attn1;
    BYTE	ksl_attn2;
    BYTE	att_dec1;
    BYTE	att_dec2;
    BYTE	sus_rel1;
    BYTE	sus_rel2;

    BYTE	wave_sel1;
    BYTE	wave_sel2;
    BYTE	fdbk_alg;
} PatchData;



const PatchData sample_patch = {
    (1<<5)|(15), /* sustain, three octaves and maj7 (times 15) */
    (1<<5)|(7),  /* sustain, two octaves and maj7 (times 7) */
    0,           /* loudest */
    0,
    0xF0, /* attack fast, release slow */
    0xFF, /* attack fast, release fast */
    0xF8, /* sustain long, release med */
    0xF8,
    0,    /* sine wave */
    3,    /* upper sawtooth */
    0x30|(3<<1)|(1) /* L&R enable; algo: parallel, feedback: 3 == PI/4 */
};

static int op_offset[]={0,1,2,8,9,10,16,17,18};

static int fnumber[]={
    0x158, /* 344 */
    0x16B,
    0x181,
    0x198,
    0x1B0,
    0x1CA,
    0x1E5,
    0x202,
    0x220,
    0x241,
    0x263,
    0x287 /* */
};

void write_adlib_reg(unsigned char reg,unsigned char value);
void write_patch(const PatchData *pd,int voice_no);
void fm_reset(void);
void play_voice(int voice_no,int octave,int note);
void stop_voice(int voice_no);
void _wa_sendcmd (unsigned int cmd);
void drum_up(void);
void nw_tone(int hertz);
void init_nw_tone(void);

static char reg_shadow[9];

/************************************************************************/

#define	STATR	0x255
#define	CMDR	0x250
#define	CMD_WE	0x80
#define	CMD_RF	0x40

void _wa_sendcmd (unsigned int cmd)
{
    int count;

    if(inb(STATR) & CMD_RF)
    {
        inw(CMDR);	//flush the port
        udelay(10);
    }

    count = 5000;	//preset timeout at 8000 loops....
    while(!(inb(STATR) & CMD_WE) && count--)   {};	//wait till bit HI

    outw(cmd, CMDR);	//output the command

    udelay(200); /* I had trouble with 100. patb */
}
void
_wa_sendcmd3(unsigned int cmd, unsigned int word1, unsigned int word2) {
    _wa_sendcmd(cmd);
    _wa_sendcmd(word1); /* misc reg */
    _wa_sendcmd(word2);
}

/************************************************************************/

void write_adlib_reg(unsigned char reg,unsigned char value)
{
    outb((unsigned char)reg,(unsigned int)ADLIB_ADDR_PORT);
    outb((unsigned char)value,(unsigned int)ADLIB_DATA_PORT);
}

/************************************************************************/

void write_patch(const PatchData *pd,int voice_no)
{
    int	addr;

    addr=op_offset[voice_no];

    write_adlib_reg(0x20+addr,pd->flags1);  /* set AM & vibrato */
    write_adlib_reg(0x20+3+addr,pd->flags2);

    write_adlib_reg(0x40+addr,pd->ksl_attn1); /* set key level */
    write_adlib_reg(0x40+3+addr,pd->ksl_attn2);

    write_adlib_reg(0x60+addr,pd->att_dec1); /* set attack/decay */
    write_adlib_reg(0x60+3+addr,pd->att_dec2);

    write_adlib_reg(0x80+addr,pd->sus_rel1); /* set sustain/release */
    write_adlib_reg(0x80+3+addr,pd->sus_rel2);

    write_adlib_reg(0xE0+addr,pd->wave_sel1); /* set waveform */
    write_adlib_reg(0xE0+3+addr,pd->wave_sel2);

    write_adlib_reg(0xC0+voice_no,pd->fdbk_alg); /* feedback algo */
}

/************************************************************************/

void fm_reset(void)
{
    int	i;

    write_adlib_reg(0x01,0x20);
    write_adlib_reg(0x08,0x20);
    write_adlib_reg(0xBD,0xC0); /* am modulation, vibrato */

    for(i=0;i<9;i++)
    {
        write_adlib_reg(0xA0,0x00); /* clear each octive/freq */
        write_adlib_reg(0xB0,0x00);
    }

    /* assure that opl3 is turned on */
    outb(5,0x38a); outb(1,0x38b);
}

/************************************************************************/

void play_voice(int voice_no,int octave,int note)
{
    unsigned char key_oct_fn;

    int	fnum=fnumber[note];

    write_adlib_reg(0xA0+voice_no,fnum&0xFF);
    key_oct_fn=0x20|((octave&0x03)<<2)|((fnum>>8)&0x03);
    reg_shadow[voice_no]=key_oct_fn;
    write_adlib_reg(0xB0+voice_no,key_oct_fn);
}

/************************************************************************/

void stop_voice(int voice_no)
{
    write_adlib_reg(0xB0+voice_no,reg_shadow[voice_no]&0x1F);
}

/************************************************************************
 this code is usable as a system beep
 starts a tone of a given freq
 stops the tone when hertz==0
 */

void nw_tone(int hertz)
{
    if (hertz != 0) {
        unsigned int fnum, octave, key_oct_fn;

        fm_reset();
        
        /* from the opl3 programming guide, the fnum
         of a tone is hertz * 2^(20-octave) / 49716 / mult_in_reg_2x
         */
        for (octave = 0; octave<7; octave++) {
            fnum = hertz * (1 << (20-octave)) / 49716 / 15;
            if (fnum<1024)
                break;
        }
        /* use voice 0: single slow attack sine wave */
	write_patch(&sample_patch,0);
	write_adlib_reg(0xA0,fnum&0xFF);
	key_oct_fn=0x20|((octave&0x03)<<2)|((fnum>>8)&0x03);
	reg_shadow[0]=key_oct_fn;
	write_adlib_reg(0xB0,key_oct_fn);
    } else {
        stop_voice(0);
        fm_reset();
    }
}

/************************************************************************/

static void _nw_nosound(unsigned long ignore_me) {
    nw_tone(0);
}

/************************************************************************/
void
init_nw_tone(void)
{
    int count;
    
    outb(8,CMDR+4); /* reset the wa */
    mdelay(5);
    outb(0,CMDR+4); /* release the reset */
    mdelay(10);
    
    count = 50000;	/* preset timeout at 5000 loops....  */
    while(!(inb(STATR) & CMD_RF) && count--)   {}	/* wait till bit HI */
    inw(CMDR); /* read back the 55AA */

    /* set the initial volume to 35% */
    _wa_sendcmd3(0x032,0x9828,0xa836);	/* misc reg */
    _wa_sendcmd3(0x032,0x4d00,0x8c80);
    _wa_sendcmd3(0x032,0x2801,0x6800);	/* enable mono out */
    _wa_sendcmd3(0x032,0x381c,0x781c);
    _wa_sendcmd3(0x032,0x1801,0x5801);	/* unmute L&R */
    _wa_sendcmd3(0x0131,0x2CCC,0x2CCC);	/* set FM volume to 35% */
}    
/************************************************************************/
/*
 this is based on the sound counter from drivers/char/vt.c
 */
void
_nw_mksound(unsigned int hz, unsigned int ticks)
{
	static struct timer_list sound_timer = { NULL, NULL, 0, 0,
						 _nw_nosound };

        del_timer(&sound_timer);

        nw_tone(hz);
        
        if (ticks) {
            sound_timer.expires = jiffies+ticks;
            add_timer(&sound_timer);
        }
	return;
}
