/*
 * Econet driver for Archimedes onboard hardware.
 *
 * (C) 1998, 1999 Phil Blundell <philb@gnu.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 */

/*
 * The Econet interface in the Archimedes is basically unchanged from
 * the original BBC B hardware (and, in fact, is precisely the same
 * card that was used on the Master 128).  Thus one shouldn't expect
 * it to be at the cutting edge of network interface design.  The 6854
 * ADLC does some of the protocol work and we have to do the rest in
 * software.  Life is hard sometimes.
 *
 * This code implements the state machine for the four-way handshake.  It
 * calls the low-level routines in ec6854_fiq.S to do the actual transfer.
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/types.h>
#include <linux/fcntl.h>
#include <linux/ptrace.h>
#include <linux/ioport.h>
#include <linux/in.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/netdevice.h>
#include <linux/skbuff.h>
#include <linux/if_arp.h>
#include <linux/if_ec.h>

#include <asm/system.h>
#include <asm/bitops.h>
#include <asm/io.h>
#include <asm/fiq.h>
#include <asm/irq.h>

#define ADLC_BASE		0x033a0000
#define ADLC_PORT_BASE		(0x80000000 | ((ADLC_BASE - IO_BASE) >> 2))

#define IRQ_EC6854		(IRQ_IMMEDIATE)
#define IRQ_EC6854_FIQ		(64 + FIQ_ECONET)

#define ADLC_CR1		0
#define ADLC_CR2		1
#define ADLC_CR3		1
#define ADLC_TXCONTINUE		2
#define ADLC_TXTERMINATE	3
#define ADLC_CR4		3
#define ADLC_SR1		0
#define ADLC_SR2		1
#define ADLC_RX			2

#define CR1_AC			0x01
#define CR1_RIE			0x02
#define CR1_TIE			0x04
#define CR1_RDSR_MODE		0x08
#define CR1_TDSR_MODE		0x10
#define CR1_DISCONTINUE		0x20
#define CR1_RXRESET		0x40
#define CR1_TXRESET		0x80

#define CR2_PSE			0x01
#define CR2_TWOBYTE		0x02
#define CR2_FLAGIDLE		0x04
#define CR2_FC			0x08
#define CR2_TXLAST		0x10
#define CR2_CLRRX		0x20
#define CR2_CLRTX		0x40
#define CR2_RTS			0x80

#define CR3_LCF			0x01
#define CR3_CEX			0x02
#define CR3_AEX			0x04
#define CR3_ZERO_ONE		0x08
#define CR3_FDSE		0x10
#define CR3_LOOP		0x20
#define CR3_TST			0x40
#define CR3_DTR			0x80

#define CR4_DFSF		0x01
#define CR4_TXWLS		0x06
#define CR4_RXWLS		0x18
#define CR4_ABT			0x20
#define CR4_ABTEX		0x40
#define CR4_NRZI		0x80

#define SR1_RDA			0x01
#define SR1_S2RQ		0x02
#define SR1_LOOP		0x04
#define SR1_FD			0x08
#define SR1_CTS			0x10
#define SR1_TXU			0x20
#define SR1_TDRA		0x40
#define SR1_IRQ			0x80

#define SR2_AP			0x01
#define SR2_FV			0x02
#define SR2_RXIDLE		0x04
#define SR2_RXABT		0x08
#define SR2_ERR			0x10
#define SR2_DCD			0x20
#define SR2_OVRN		0x40
#define SR2_RDA			0x80

#define CR2val			(CR2_PSE | CR2_TWOBYTE)

enum ec_state {
        IDLE,
	SEND_SCOUT_ACK,
        RECEIVE_DATA,
        SEND_FINAL_ACK
};

struct ec_private {
	int fiq_claimed;
	int fiq_nr;
	unsigned char frame_buffer[16];
	int my_station_nr;
	enum ec_state state;
	struct sk_buff *skb;
};

extern void ec_fiq_code, ec_fiq_code_end, ec_fiq_send_frame,
	ec_fiq_receive;

static void
install_fiq_handler(struct device *dev)
{
	struct ec_private *priv = (struct ec_private *)dev->priv;
	struct pt_regs regs;
	set_fiq_handler(&ec_fiq_code, &ec_fiq_code_end - &ec_fiq_code);
	regs.ARM_r8 = ADLC_BASE;
	regs.ARM_r9 = (unsigned long)&priv->frame_buffer[0];
	regs.ARM_r10 = regs.ARM_r9 + 14;
	regs.ARM_ip = 0;
	regs.ARM_sp = priv->my_station_nr;
	priv->fiq_claimed = 1;
	priv->state = IDLE;
	set_fiq_regs(&regs);
	outb(CR1_RIE, dev->base_addr + ADLC_CR1);
	outb(CR2val, dev->base_addr + ADLC_CR2);
	enable_irq(priv->fiq_nr);
}

/*
 * The kernel calls us here when someone else wants to take the FIQ away,
 * or when we can have it back.
 */

static int 
ec_fiq_op(void *dev_id, int relinquish)
{
	struct device *dev = (struct device *)dev_id;
	struct ec_private *priv = (struct ec_private *)dev->priv;
	if (relinquish) {
		outb(0, dev->base_addr + ADLC_CR1);
		disable_irq(priv->fiq_nr);
		priv->fiq_claimed = 0;
	} else {
		install_fiq_handler(dev);
	}
	return 0;
}

static struct fiq_handler ec_fh = { NULL, "econet", ec_fiq_op, NULL };

/*
 * This hangs on the FIQ downgrade interrupt.  The assembler code causes
 * us to be called here when something "interesting" happens.
 */

static void
ec6854_irq(int irq, void *dev_id, struct pt_regs *regs)
{
	struct device *dev = (struct device *)dev_id;
	struct ec_private *priv = (struct ec_private *)dev->priv;
	struct pt_regs fiq_regs;
	struct ec_framehdr *frm;
	struct sk_buff *skb, *newskb;
	int stn, net;

	get_fiq_regs(&fiq_regs);
	switch (priv->state) {
	case IDLE:
		/* The FIQ code saw an incoming scout that seemed to be
		   for us.  ADLC is in time fill on the wire. */
		/* First check for status bits that mean the packet
		   is junk. */
		if (fiq_regs.ARM_ip 
		    & (SR2_OVRN | SR2_ERR | SR2_RXABT | SR2_RXIDLE)) {
			printk("%s: received status %02x\n", dev->name,
			       fiq_regs.ARM_ip);
			goto reset;
		}
		if (fiq_regs.ARM_ip == 0) {
			/* The buffer we gave was too short.  A scout can
			   only be six bytes legitimately so this means
			   something is out of sync.  Give up and go back
			   to idling.  */
			printk("%s: receive buffer screwed.\n", dev->name);
			goto reset;
		}
		frm = (struct ec_framehdr *)&priv->frame_buffer[0];
		stn = frm->src_stn;
		net = frm->src_net;
		/* Do we really want the packet? */
		if (ec_listening_socket(frm->port, stn, net)) {
			/* Construct a scout acknowledge and send it out. */
			frm->src_stn = priv->my_station_nr;
			frm->src_net = 0;
			frm->dst_stn = stn;
			frm->dst_net = net;
			priv->state = SEND_SCOUT_ACK;
			fiq_regs.ARM_r9 = 
				(unsigned long)&priv->frame_buffer[2];
			fiq_regs.ARM_r10 = fiq_regs.ARM_r9 + 2;
			fiq_regs.ARM_ip = 0;
			set_fiq_regs(&fiq_regs);
			*((void **)0x20) = &ec_fiq_send_frame;
			outb(priv->frame_buffer[0], 
			     dev->base_addr + ADLC_TXCONTINUE);
			outb(priv->frame_buffer[1], 
			     dev->base_addr + ADLC_TXCONTINUE);
			/* Enable Tx ints, leave receiver in reset */
			outb(CR1_RXRESET | CR1_TIE, 
			     dev->base_addr + ADLC_CR1);
			break;
		}
		printk("%s: No takers for %02x.%02x port %02x\n",
		       dev->name, net, stn, frm->port);
		/* Higher level code decided nobody was listening for this
		   packet.  Chuck it and go back to the beginning. */
		goto reset;

	case SEND_SCOUT_ACK:
		/* Scout acknowledge has gone.  FIQ code has taken the
                   ADLC off the wire and enabled the receiver.  We
                   just need to turn around and get ready for the next
                   frame. */
		priv->state = RECEIVE_DATA;
		fiq_regs.ARM_r9 = (unsigned long)priv->skb->data;
		fiq_regs.ARM_r10 = fiq_regs.ARM_r9 + 8191;
		fiq_regs.ARM_ip = 0;
		*((void **)0x20) = &ec_fiq_receive;
		set_fiq_regs(&fiq_regs);
		outb(CR1_RIE, dev->base_addr + ADLC_CR1);
		outb(CR2val, dev->base_addr + ADLC_CR2);
		break;

	case RECEIVE_DATA:
		/* First check for status bits that mean the packet
		   is junk. */
		if (fiq_regs.ARM_ip 
		    & (SR2_OVRN | SR2_ERR | SR2_RXABT | SR2_RXIDLE)) {
			printk("%s: received status %02x\n", dev->name,
			       fiq_regs.ARM_ip);
			goto reset;
		}
		if (fiq_regs.ARM_ip == 0) {
			/* The buffer we gave was too short.  In
			   theory someone could have sent us a
			   jumbogram but more likely this means the
			   driver is buggy.  Give up and go back to
			   idling.  */
			printk("%s: receive buffer screwed.\n", dev->name);
			goto reset;
		}
		skb = priv->skb;
		skb_put(skb, fiq_regs.ARM_r9 - (unsigned long)skb->data);
		skb->mac.raw = skb->data;
		/* Construct a final acknowledge and send it out. */
		frm = (struct ec_framehdr *)skb->data;
		stn = frm->src_stn;
		net = frm->src_net;
		frm->src_stn = priv->my_station_nr;
		frm->src_net = 0;
		frm->dst_stn = stn;
		frm->dst_net = net;
		priv->state = SEND_FINAL_ACK;
		fiq_regs.ARM_r9 = 
			(unsigned long)&priv->frame_buffer[2];
		fiq_regs.ARM_r10 = fiq_regs.ARM_r9 + 2;
		fiq_regs.ARM_ip = 0;
		set_fiq_regs(&fiq_regs);
		*((void **)0x20) = &ec_fiq_send_frame;
		outb(priv->frame_buffer[0], 
		     dev->base_addr + ADLC_TXCONTINUE);
		outb(priv->frame_buffer[1], 
		     dev->base_addr + ADLC_TXCONTINUE);
		outb(CR1_RXRESET | CR1_TIE, 
		     dev->base_addr + ADLC_CR1);
		break;

	case SEND_FINAL_ACK:
		/* The handshake is complete.  Send the packet up the
                   network stack.  In theory we could have done this
                   earlier since this phase doesn't bring any useful
                   data, but we defer it to here since the turnaround
                   between data and acknowledge is time critical.  */
		skb = priv->skb;
	        newskb = dev_alloc_skb(8192);
		if (newskb) {
			skb = priv->skb;
			skb->dev = dev;
			skb->protocol = htons(ETH_P_ECONET);
			netif_rx(skb);
			priv->skb = newskb;
		} else 
			printk("%s: memory squeeze.\n", dev->name);
		goto reset;
	}
	return;

 reset:
	/* Go back to the idle state */
	priv->state = IDLE;
	fiq_regs.ARM_r9 = (unsigned long)&priv->frame_buffer[0];
	fiq_regs.ARM_r10 = fiq_regs.ARM_r9 + 14;
	fiq_regs.ARM_ip = 0;
	*((void **)0x20) = &ec_fiq_receive;
	set_fiq_regs(&fiq_regs);
	outb(CR1_RIE, dev->base_addr + ADLC_CR1);
	outb(CR2val, dev->base_addr + ADLC_CR2);
}

static inline int
clock_detected(struct device *dev)
{
	return !(inb(dev->base_addr + ADLC_SR2) & SR2_DCD);
}

static int
ec6854_open(struct device *dev)
{
	struct ec_private *priv = (struct ec_private *)dev->priv;
	struct ec_device *edev = edev_get(dev);

	if (!edev) {
		printk(KERN_ERR "%s: no protocol address given yet.\n",
		       dev->name);
		return -EINVAL;
	}

	priv->my_station_nr = edev->station;

	if (request_irq(dev->interrupt, ec6854_irq, 0, "econet", dev))
		return -EAGAIN;

	priv->skb = dev_alloc_skb(8192);
	if (priv->skb == NULL) {
		printk("%s: memory squeeze.\n", dev->name);
		return -ENOMEM;
	}

	outb(CR1_AC, dev->base_addr + ADLC_CR1);
	outb(0, dev->base_addr + ADLC_CR3);
	outb(CR4_TXWLS | CR4_RXWLS, dev->base_addr + ADLC_CR4);
	outb(0, dev->base_addr + ADLC_CR1);
	outb(CR2val, dev->base_addr + ADLC_CR2);

	if (claim_fiq(&ec_fh)) {
		printk("ec0: couldn't claim FIQ.\n");
		kfree_skb(priv->skb);
		return -EAGAIN;
	}

	install_fiq_handler(dev);

	if (!clock_detected(dev))
		printk(KERN_WARNING "%s: warning, no clock.\n", dev->name);

	outb(CR2val | CR2_CLRRX, dev->base_addr + ADLC_CR2);

	return 0;
}

static int
ec6854_stop(struct device *dev)
{
	struct ec_private *priv = (struct ec_private *)dev->priv;

	printk("%s: status %02x, %02x\n", dev->name,
	       inb(dev->base_addr + ADLC_SR1),
	       inb(dev->base_addr + ADLC_SR2));

	outb(CR1_RXRESET | CR1_TXRESET, dev->base_addr);
	outb(CR2val, dev->base_addr + ADLC_CR2);
	disable_irq(priv->fiq_nr);
	kfree_skb(priv->skb);
	priv->fiq_claimed = 0;
	release_fiq(&ec_fh);
	free_irq(dev->interrupt, dev);
	return 0;
}

int __init
ec6854_probe(struct device *dev)
{
	unsigned long base = ADLC_PORT_BASE;
	struct ec_private *priv = kmalloc(sizeof(struct ec_private), 
					  GFP_KERNEL);

	if (!priv) {
		printk("ec6854: memory squeeze.\n");
		return -ENOMEM;
	}

	if (check_region(base, 16)) {
		kfree(priv);
		return -ENODEV;
	}
	
	request_region(base, 16, "econet/6854");

	ec_fh.dev_id = dev;
	
	printk(KERN_INFO "%s: MC6854 Econet hardware detected\n", dev->name);

	dev->priv = priv;
	dev->interrupt = IRQ_EC6854;
	dev->base_addr = base;
	dev->mem_start = ADLC_BASE;
	dev->type = ARPHRD_ECONET;
	dev->hard_header_len = 6;
	dev->addr_len = 2;
	dev->tx_queue_len = 10;
	dev->mtu = 8192;
	memset(dev->broadcast, 0xFF, 2);
	dev_init_buffers(dev);

	dev->open = ec6854_open;
	dev->stop = ec6854_stop;

	priv->fiq_nr = IRQ_EC6854_FIQ;
	priv->fiq_claimed = 0;

	outb(CR1_RXRESET | CR1_TXRESET, dev->base_addr);

	return 0;
}
