/*
 * Generic 21285 Footbridge flash ROM driver
 * Copyright (C) 1998 Philip Blundell <pb@nexus.co.uk>
 *
 * Derived from drivers/sbus/char/flash.c:  
 * Copyright (C) 1997  Eddie C. Dost  (ecd@skynet.be)
 */

#include <linux/module.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/miscdevice.h>
#include <linux/malloc.h>
#include <linux/fcntl.h>
#include <linux/poll.h>
#include <linux/init.h>

#include <asm/system.h>
#include <asm/uaccess.h>
#include <asm/pgtable.h>
#include <asm/io.h>
#include <asm/dec21285.h>

static struct {
	unsigned long base;
	unsigned long virt_base;
	unsigned long size;
	unsigned long busy;
} flash;

#define FLASH_MINOR	152

static int
flash_mmap(struct file *file, struct vm_area_struct *vma)
{
	unsigned long addr;
	unsigned long size;

	if (vma->vm_offset & ~(PAGE_MASK))
		return -ENXIO;

	addr = flash.base;
	size = flash.size;

	if (vma->vm_offset > size)
		return -ENXIO;
	addr += vma->vm_offset;

	if (vma->vm_end - (vma->vm_start + vma->vm_offset) > size)
		size = vma->vm_end - (vma->vm_start + vma->vm_offset);

	vma->vm_flags |= (VM_SHM | VM_LOCKED);

	pgprot_val(vma->vm_page_prot) &= ~(PTE_CACHEABLE | PTE_BUFFERABLE);
		
	if (remap_page_range(vma->vm_start, addr, size, vma->vm_page_prot))
		return -EAGAIN;
		
	vma->vm_file = file;
	file->f_count++;
	return 0;
}

static long long
flash_llseek(struct file *file, long long offset, int origin)
{
	switch (origin) {
		case 0:
			file->f_pos = offset;
			break;
		case 1:
			file->f_pos += offset;
			if (file->f_pos > flash.size)
				file->f_pos = flash.size;
			break;
		case 2:
			file->f_pos = flash.size;
			break;
		default:
			return -EINVAL;
	}
	return file->f_pos;
}

static ssize_t
flash_read(struct file * file, char * buf,
	   size_t count, loff_t *ppos)
{
	unsigned long p = file->f_pos;
	
	if (count > flash.size - p)
		count = flash.size - p;

	if (copy_to_user(buf, (void *)flash.virt_base + p, count) < 0)
		return -EFAULT;

	file->f_pos += count;
	return count;
}

static int
flash_open(struct inode *inode, struct file *file)
{
	if (test_and_set_bit(0, (void *)&flash.busy) != 0)
		return -EBUSY;

	MOD_INC_USE_COUNT;
	return 0;
}

static int
flash_release(struct inode *inode, struct file *file)
{
	MOD_DEC_USE_COUNT;
	flash.busy = 0;
	return 0;
}

static struct file_operations flash_fops = {
	flash_llseek,
	flash_read,
	NULL,		/* no write to the Flash, use mmap
			 * and play flash dependent tricks.
			 */
	NULL,		/* readdir */
	NULL,		/* poll */
	NULL,		/* ioctl */
	flash_mmap,
	flash_open,
	NULL,		/* flush */
	flash_release
};

static struct miscdevice flash_dev = { FLASH_MINOR, "flash", &flash_fops };

EXPORT_NO_SYMBOLS;

#ifdef MODULE
int init_module(void)
#else
__initfunc(int flash_init(void))
#endif
{
	int err;

	flash.base = DC21285_FLASH;
	flash.size = 4*1024*1024 / 8;
	/* For now we use a fixed mapping set up by the MM init code.
	   This should probably go away and be replaced by a call to
	   ioremap */
	flash.virt_base = 0xd8000000;

	err = misc_register(&flash_dev);
	if (err) {
		printk(KERN_ERR "flash: unable to get misc minor\n");
		return err;
	}

	return 0;
}

#ifdef MODULE
void cleanup_module(void)
{
	misc_deregister(&flash_dev);
}
#endif
