/*
 * Driver for Philips SAB3036 "CITAC" tuner control chip.
 *
 * Author: Phil Blundell <philb@gnu.org>
 *
 * The SAB3036 is just about different enough from the chips that
 * tuner.c copes with to make it not worth the effort to crowbar
 * the support into that file.  So instead we have a separate driver. 
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/malloc.h>
#include <linux/version.h>

#include <linux/i2c.h>
#include <linux/videodev.h>

#include "tuner.h"

static int debug =  0; /* insmod parameter */

MODULE_PARM(debug,"i");

struct tuner
{
	struct i2c_bus   *bus;     /* where is our chip */
	int               addr;
};

/* ---------------------------------------------------------------------- */

static int 
tuner_getstatus (struct tuner *t)
{
	return i2c_read(t->bus, t->addr+1);
}

#define TUNER_FL        0x80

static int 
tuner_islocked (struct tuner *t)
{
        return (tuner_getstatus(t) & TUNER_FL);
}

/* ---------------------------------------------------------------------- */

static void 
set_tv_freq(struct tuner *t, int freq)
{
	u16 div = ((freq * 20) / 16);
	unsigned long give_up = jiffies + HZ;
	LOCK_FLAGS;

	if (debug)
		printk(KERN_DEBUG "tuner: setting frequency %dMHz, divisor %x\n", freq / 16, div);
	
	LOCK_I2C_BUS(t->bus);
	
	/* Select high tuning current */
	if (i2c_write(t->bus, t->addr, 0x29, 0x3e, 1) < 0) 
		printk("tuner: i2c i/o error\n");
	
	if (i2c_write(t->bus, t->addr, 0x80 | ((div>>8) & 0x7f), 
		      div&0xff, 1) < 0) 
		printk("tuner: i2c i/o error\n");
	
	while (!tuner_islocked(t) && time_before(jiffies, give_up))
		schedule();
	       
	if (!tuner_islocked(t))
		printk(KERN_WARNING "tuner: failed to achieve PLL lock\n");
	
	/* Select low tuning current and engage AFC */
	if (i2c_write(t->bus, t->addr, 0x29, 0xb2, 1) < 0) 
		printk("tuner: i2c i/o error\n");

	if (debug)
		printk(KERN_DEBUG "tuner: status %02x\n", tuner_getstatus(t));
	
	UNLOCK_I2C_BUS(t->bus);
}

/* ---------------------------------------------------------------------- */

static int 
tuner_attach(struct i2c_device *device)
{
	struct tuner *t;
	LOCK_FLAGS;

	/*
	 *	For now we only try and attach these tuners to the parallel
	 *	bus. This same module will however work different species
	 *	of card using these chips. Just change the constraints
	 *	(i2c doesn't have a totally clash free 'address' space)
	 */
	
	if(device->bus->id != I2C_BUSID_PARPORT)
		return -EINVAL;
		
	t = kmalloc(sizeof(struct tuner), GFP_KERNEL);
	if (t == NULL)
		return -ENOMEM;
	memset(t, 0, sizeof(struct tuner));
	strcpy(device->name, "tuner");

	t->bus  = device->bus;
	t->addr = device->addr;

	printk("tuner: SAB3036 found, status %02x\n", tuner_getstatus(t));

	LOCK_I2C_BUS(t->bus);

	if (i2c_write(t->bus, t->addr, 0x29, 0x32, 1) < 0) 
		printk("tuner: i2c i/o error\n");

	if (i2c_write(t->bus, t->addr, 0x2a, 0x0, 1) < 0) 
		printk("tuner: i2c i/o error\n");

	if (i2c_write(t->bus, t->addr, 0x2b, 0x0, 1) < 0) 
		printk("tuner: i2c i/o error\n");

	UNLOCK_I2C_BUS(t->bus);

	device->data = t;
	
	MOD_INC_USE_COUNT;
	return 0;
}

static int 
tuner_detach(struct i2c_device *device)
{
	struct tuner *t = (struct tuner*)device->data;
	kfree(t);
	MOD_DEC_USE_COUNT;
	return 0;
}

static int 
tuner_command(struct i2c_device *device,
	      unsigned int cmd, void *arg)
{
	struct tuner *t = (struct tuner*)device->data;
	int *iarg = (int*)arg;

	switch (cmd) 
	{
		case TUNER_SET_TVFREQ:
			set_tv_freq(t,*iarg);
			break;
	    
		default:
			return -EINVAL;
	}
	return 0;
}

/* ----------------------------------------------------------------------- */

struct i2c_driver 
i2c_driver_tuner = 
{
	"sab3036",                    /* name       */
	I2C_DRIVERID_TUNER,           /* ID         */
	0xc0, 0xc1,                   /* addr range */

	tuner_attach,
	tuner_detach,
	tuner_command
};

EXPORT_NO_SYMBOLS;

#ifdef MODULE
int init_module(void)
#else
int i2c_tuner_init(void)
#endif
{
	i2c_register_driver(&i2c_driver_tuner);
	return 0;
}

#ifdef MODULE
void cleanup_module(void)
{
	i2c_unregister_driver(&i2c_driver_tuner);
}
#endif
