/*
 *  linux/fs/adfs/map.c
 *
 * Copyright (C) 1997 Russell King
 */

#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/adfs_fs.h>

#include "adfs.h"

/*
 * return the map bit offset of the fragment frag_id in
 * the zone dm.
 * Note that the loop is optimised for best asm code -
 * look at the output of:
 *  gcc -D__KERNEL__ -O2 -I../../include -o - -S map.c|less
 */
static int
lookup_zone(const struct adfs_discmap *dm, const unsigned int idlen,
	    const unsigned int frag_id, unsigned int *offset)
{
	const unsigned int mapsize = dm->dm_endbit;
	const unsigned int idmask = (1 << idlen) - 1;
	unsigned long *map = ((unsigned long *)dm->dm_bh->b_data) + 1;
	unsigned int start = dm->dm_startbit;
	unsigned int mapptr;

	do {
		unsigned long frag;

		/*
		 * get fragment id
		 */
		__asm__("@ get fragment id start");
		{
			unsigned long v2;
			unsigned int tmp;


			tmp  = start >> 5;

			frag = map[tmp];
			v2   = map[tmp + 1];

			tmp  = (start & 31);

			frag = (frag >> tmp) | (v2 << (32 - tmp));

			frag &= idmask;

		}
		__asm__("@ get fragment id end");

		mapptr = start + idlen;

		/*
		 * find end of fragment
		 */
		__asm__("@ find end of fragment start");
		{
			unsigned long v2;

			while ((v2 = map[mapptr >> 5] >> (mapptr & 31)) == 0) {
				mapptr = (mapptr & ~31) + 32;
				if (mapptr >= mapsize)
					goto error;
			}

			mapptr += 1 + ffz(~v2);
		}
		__asm__("@ find end of fragment end");

		if (frag == frag_id)
			goto found;
again:
		start = mapptr;
	} while (mapptr < mapsize);

error:
	return -1;

found:
	{
		int length = mapptr - start;
		if (*offset >= length) {
			*offset -= length;
			goto again;
		}
	}
	return start + *offset;
}

static int
scan_map(struct adfs_sb_info *asb, unsigned int zone,
	 const unsigned int frag_id, unsigned int mapoff)
{
	const unsigned int idlen = asb->s_idlen;
	struct adfs_discmap *dm, *dm_end;
	int result;

	dm       = asb->s_map + zone;
	zone     = asb->s_map_size;
	dm_end   = asb->s_map + zone;

	do {
		result = lookup_zone(dm, idlen, frag_id, &mapoff);

		if (result != -1)
			goto found;

		dm ++;
		if (dm == dm_end)
			dm = asb->s_map;
	} while (--zone > 0);

	return -1;
found:
	result -= dm->dm_startbit;
	result += dm->dm_startblk;

	return result;
}

int adfs_map_lookup(struct super_block *sb, int frag_id, int offset)
{
	struct adfs_sb_info *asb = &sb->u.adfs_sb;
	unsigned int zone, mapoff;
	int result;

	zone   = frag_id / asb->s_ids_per_zone;

	if (zone >= asb->s_map_size)
		goto bad_fragment;

	/* Convert sector offset to map offset */
	mapoff = signed_asl(offset, -asb->s_map2blk);

	result = scan_map(asb, zone, frag_id, mapoff);

	if (result > 0) {
		unsigned int secoff;

		/* Calculate sector offset into map block */
		secoff = offset - signed_asl(mapoff, asb->s_map2blk);
		return secoff + signed_asl(result, asb->s_map2blk);
	}

	adfs_error(sb, "adfs_map_lookup",
		   "fragment %X at offset %d not found in map",
		    frag_id, offset);
	return 0;

bad_fragment:
	adfs_error(sb, "adfs_map_lookup",
		   "fragment %X is invalid (zone = %d, max = %d)",
		    frag_id, zone, asb->s_map_size);
	return 0;
}
