
/*
 * This kernel module, designed to operate on a forwarding or masquerading
 * Linux 2.2 or 2.[3-4] gateway that has attached tunnels or encapsulation
 * interfaces with reduced packet sizes (like PPPoE) connected to networks
 * with ICMP blackholes (such as Bell Canada's Sympatico HSE), prevents
 * Path MTU Discovery failures (causing a TCP connection to hang as soon
 * as the remote host tries to send full packets) by patching
 * the MSS option in TCP SYN packets down to the safest value
 * according to the outgoing device's MTU.
 *
 * Idea and implementation by Marc Boucher <marc@mbsi.ca>
 * Linux 2.2 CONFIG_FIREWALL support done under contract for Rebel.com.
 * Check out the Netwinder/OfficeServer at http://www.rebel.com/netwinder/ !
 *
 * Copyright (C) 1999, 2000 Marc Boucher and Rebel.com. All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation, and
 * other materials related to such distribution and use acknowledge
 * that the software was developed by Marc Boucher and Rebel.com.
 * The name of the authors may not be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * $Id: mssclampfw.c,v 1.1.2.3 2000/01/12 15:54:22 ralphs Exp $
 */

#define MSSCLAMPFW_VERSION_MAJOR 1
#define MSSCLAMPFW_VERSION_MINOR 2

#include <linux/config.h>

#if !defined(CONFIG_FIREWALL) && !defined(CONFIG_NETFILTER)
#error You must enable CONFIG_FIREWALL or CONFIG_NETFILTER in your kernel config
#endif

#include <linux/version.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/file.h>
#include <linux/malloc.h>
#include <linux/miscdevice.h>
#include <linux/poll.h>
#include <linux/skbuff.h>
#include <linux/in.h>
#include <linux/ip.h>
#include <linux/init.h>
#include <linux/proc_fs.h>
#include <net/protocol.h>
#include <net/tcp.h>
#include <net/udp.h>

#ifdef CONFIG_NETFILTER
#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>
#else
#include <linux/firewall.h>
#endif

#ifndef IP_PARTS
#define IP_PARTS_NATIVE(n)			\
(unsigned int)((n)>>24)&0xFF,			\
(unsigned int)((n)>>16)&0xFF,			\
(unsigned int)((n)>>8)&0xFF,			\
(unsigned int)((n)&0xFF)

#define IP_PARTS(n) IP_PARTS_NATIVE(ntohl(n))
#endif

#ifdef CONFIG_NETFILTER
#define MSSCLRETVAL NF_ACCEPT
static unsigned int
mssClampFwForwardHook (unsigned int hook,
				struct sk_buff **pskb,
				const struct net_device *indev,
				const struct net_device *outdev,
				int (*okfn)(struct sk_buff *))
#else
#define MSSCLRETVAL FW_SKIP
static int
mssClampFwForwardCheck (struct firewall_ops *this,
				int pf,
				struct device *outdev,
				void *phdr,
				void *arg,
				struct sk_buff **pskb)
#endif
{
	struct iphdr *iph;
	struct tcphdr *tcph;
	u_int16_t tcplen;
	unsigned int i;
	u_int8_t *opt;
	int mss_clamp;

#ifdef CONFIG_NETFILTER
	iph = (*pskb)->nh.iph;
#else
	if(pf != PF_INET)  {
		return MSSCLRETVAL;
	}

	iph = (struct iphdr *)phdr;
#endif

	if(iph->protocol != IPPROTO_TCP) {
		return MSSCLRETVAL;
	}

	if(ntohs(iph->frag_off) & IP_OFFSET) {
		return MSSCLRETVAL;
	}

	tcplen = (*pskb)->len - iph->ihl*4;

	tcph = (void *)iph + iph->ihl*4;

	if(((*pskb)->len < (iph->ihl*4 + sizeof(struct tcphdr))) || (tcplen < tcph->doff*4)) {
		printk(KERN_ERR "mssclampfw: skb too short (%d bytes)\n", (*pskb)->len);
		return MSSCLRETVAL;
	}

	if(!(tcph->syn && !tcph->ack && !tcph->rst)) {
		return MSSCLRETVAL;
	}

        /* Checksum invalid?  Ignore. */
        if (tcp_v4_check(tcph, tcplen, iph->saddr, iph->daddr,
                         csum_partial((char *)tcph, tcplen, 0))) {
                printk(KERN_ERR "mssclampfw: bad csum: %p %u %u.%u.%u.%u %u.%u.%u.%u\n",
                         tcph, tcplen, IP_PARTS(iph->saddr),
                         IP_PARTS(iph->daddr));
		return MSSCLRETVAL;
        }

 	i = sizeof(struct tcphdr);
 	opt = (u_int8_t *)tcph;

	mss_clamp = ((outdev && outdev->mtu) ? outdev->mtu : 1500) - sizeof(struct iphdr) - sizeof(struct tcphdr);

	while (i < tcph->doff*4) {
		if ((opt[i] == TCPOPT_MSS) &&
				((tcph->doff*4 - i) >= TCPOLEN_MSS) &&
				(opt[i+1] == TCPOLEN_MSS)) {

			u_int16_t oldmss;
			u_int32_t csumdiffs[2];

			oldmss = (opt[i+2] << 8) | opt[i+3];
			if(oldmss < mss_clamp) {
				/* already ok */
				return MSSCLRETVAL;
			}

			opt[i+2] = (mss_clamp & 0xff00) >> 8;
			opt[i+3] = (mss_clamp & 0x00ff);

			/*
 			* Recalculate checksum using differences between
 			* changed datum.
 			*/

			csumdiffs[0] = htons(~oldmss);
			csumdiffs[1] = htons(mss_clamp);

			tcph->check = csum_fold(csum_partial((char *)csumdiffs, sizeof(csumdiffs), tcph->check^0xFFFF));

#if 0 || defined(MSSCLAMPFW_DEBUG)
			printk(KERN_INFO "mssclampfw: %u.%u.%u.%u:%u->%u.%u.%u.%u:%u changed TCP MSS option from %u to %u [outdev=%s mtu=%d]\n", IP_PARTS(iph->saddr), ntohs(tcph->source), IP_PARTS(iph->daddr), ntohs(tcph->dest), oldmss, mss_clamp, (outdev && outdev->name) ? outdev->name : "NULL", outdev ? outdev->mtu : 0 );
#endif
			goto retmodified;
		}
		if (opt[i] < 2) i++;
		else i += opt[i+1]?:1;
	}

	if (skb_tailroom((*pskb)) < TCPOLEN_MSS) {
		printk(KERN_ERR "mssclampfw: not enough tailroom in skb to add TCP MSS option\n");
		return MSSCLRETVAL;
	}

	/*
	 * MSS Option not found ? add it..
	 */

	skb_put((*pskb), TCPOLEN_MSS);

 	opt = (u_int8_t *)tcph + sizeof(struct tcphdr);
	memmove(opt + TCPOLEN_MSS, opt, tcplen - sizeof(struct tcphdr));

	opt[0] = TCPOPT_MSS;
	opt[1] = TCPOLEN_MSS;
	opt[2] = (mss_clamp & 0xff00) >> 8;
	opt[3] = (mss_clamp & 0x00ff);

	tcph->doff += TCPOLEN_MSS/4;

	tcplen = (*pskb)->len - iph->ihl*4;

	iph->tot_len = htons(ntohs(iph->tot_len) + TCPOLEN_MSS);

	/* recalc IP checksum */
	ip_send_check(iph);

	/* recalc TCP checksum */
	if(tcplen - tcph->doff*4) {
		(*pskb)->csum = csum_partial((void *)tcph + tcph->doff*4, tcplen - tcph->doff*4, 0);
	} else {
		/* shouldn't csum_partial return 0 in this case? */
		(*pskb)->csum = 0;
	}
	tcph->check = 0;
	tcph->check = tcp_v4_check(tcph, tcplen, iph->saddr, iph->daddr,
		csum_partial((char *)tcph, tcph->doff*4,
		(*pskb)->csum));

#if 0 || defined(MSSCLAMPFW_DEBUG)
	printk(KERN_DEBUG "mssclampfw: %u.%u.%u.%u:%u->%u.%u.%u.%u:%u added TCP MSS option = %u\n", IP_PARTS(iph->saddr), ntohs(tcph->source), IP_PARTS(iph->daddr), ntohs(tcph->dest), mss_clamp);
#endif

retmodified:
#if 0 || defined(MSSCLAMPFW_DEBUG)
        if (tcp_v4_check(tcph, tcplen, iph->saddr, iph->daddr,
			csum_partial((char *)tcph, tcplen, 0))) {
		printk(KERN_ERR "mssclampfw: bad new csum: %p %u %u.%u.%u.%u %u.%u.%u.%u\n", tcph, tcplen, IP_PARTS(iph->saddr), IP_PARTS(iph->daddr));
        }
#endif

	(*pskb)->ip_summed = CHECKSUM_NONE;
#ifdef CONFIG_NETFILTER
	(*pskb)->nfcache |= NFC_UNKNOWN | NFC_ALTERED;
#endif

	return MSSCLRETVAL;
}

#ifdef CONFIG_NETFILTER
static struct nf_hook_ops mssclampfw_ops = {
	{ NULL, NULL },
	mssClampFwForwardHook,
	NULL,
	PF_INET,
	NF_IP_FORWARD,
	INT_MAX
};

#else

static int
mssClampFwSkip (struct firewall_ops *this, int pf, struct device *dev,
			void *phdr, void *arg, struct sk_buff **pskb)
{
	return FW_SKIP;
}

static struct firewall_ops mssclampfw_ops = {
	NULL, /* next */
	mssClampFwForwardCheck,
	mssClampFwSkip,
	mssClampFwSkip,
	PF_INET,
	INT_MAX /* highest priority because we want to be first in the chain */
};
#endif /* CONFIG_NETFILTER */

#ifdef MODULE
EXPORT_NO_SYMBOLS;
int init_module(void)
#else
__initfunc(int mssclampfw_init(void))
#endif
{
	int res;

#ifdef CONFIG_NETFILTER
	res = nf_register_hook(&mssclampfw_ops);
#else
	res = register_firewall(PF_INET, &mssclampfw_ops);
#endif
	if(res) {
		printk(KERN_ERR "mssclampfw_init: failed\n");
		return res;
	}

	printk(KERN_INFO "mssclampfw v%d.%d (written by Marc Boucher <marc@mbsi.ca>. Thanks to Rebel.com)\n", MSSCLAMPFW_VERSION_MAJOR, MSSCLAMPFW_VERSION_MINOR);

	return 0;
}

#ifdef MODULE
void cleanup_module(void)
{
#ifdef CONFIG_NETFILTER
	nf_unregister_hook(&mssclampfw_ops);
#else
	unregister_firewall(PF_INET, &mssclampfw_ops);
#endif
	printk(KERN_INFO "unregistered mssclampfw\n");
} /* cleanup_module */
#endif /* MODULE */

/*
 * $Log: mssclampfw.c,v $
 * Revision 1.1.2.3  2000/01/12 15:54:22  ralphs
 * Update to mssclamp-0.2, fixing an issue with adding MSS to packets
 * that don't already have the option set in them.  Also revert the
 * formatting changes made earlier (easier to sync with Marc Boucher).
 *
 * Revision 1.3  2000/01/08 08:29:54  marcbou
 * Fixed byte-ordering bug with iph->tot_len when adding MSS option
 *
 * Revision 1.2  1999/12/21 15:34:12  marcbou
 * Implemented NETFILTER support for 2.3/2.4
 * Cleaned things up
 *
 * Revision 1.1  1999/12/14 04:26:19  marcbou
 * Initial revision
 *
 */
